//__INSERT_LICENSE__
// $Id: ralea.cpp,v 1.8 2004/04/05 16:14:00 mstorti Exp $

/*
  COMIENZO DE DESCRIPCION

  Usando las operaciones de Lista, escribir una funcion `void
  random_shufle(list<int> &L)' que, dada una lista de enteros `L'
  reordena sus elementos en forma aleatoria.  Se sugiere el siguiente
  algoritmo: Usando una lista auxiliar Q, se van generando numeros
  enteros desde 0 a `length(L)-1'. Se extrae el elemento `j-esimo' de
  L y se inserta en Q. Finalmente, se vuelven a pasar todos los
  elementos de Q a L. 
  [Ejercicio tomado en el Ex\'amen Final del 05/07/01]
  keywords: lista
  
  FIN DE DESCRIPCION
*/

#include <list>
#include <iostream>
#include <cassert>
#include <map>
#include "./util.h"

using namespace std;

void random_shuffle(list<int> &L) {
  // Cuenta el numero de elementos en L
  int n=0;
  list<int>::iterator p = L.begin();
  list<int> Q;
  while (p++!=L.end()) n++;
  // En cada iteracion del lazo se saca un elemento
  // al azar y se inserta en Q
  for (int m=n; m>0; m--) {
    // A esta altura la lista tiene `m' elementos
    // asi que generamos un entero entre 0 y `n-1'
    int k = irand(m);
    // Nos posicionamos en el elemento k en L
    p = L.begin();
    for (int j=0; j<m-1; j++) p++;
    // Inserta el elemento `k' de L en al fin de
    // Q y lo elimina de L
    Q.insert(Q.end(),*p);
    L.erase(p);
  }
  // Vuelve a copiar todos los elementos de Q a L.
  list<int>::iterator q = Q.begin();
  while (q!=Q.end()) {
    L.insert(L.end(),*q);
    q = Q.erase(q);
  }
}

void random_shuffle2(list<int> &L) {
  // Esta version es bastante mas sofisticada.
  // in-place y O(n log n). Se basa en 
  // dividir la lista en dos sublistas, aplicarles
  // el random_shuffle2 a cada una de ellas y hacer
  // un `random merge' de las dos sublistas. 

  // Dividimos la lista en dos de longitud random. 
  // Para eso vamos tirando la moneda y dependiendo
  // de si es cara o ceca la ponemos en L1 o L2. 
  // n1 y n2 contienen el numero de elementos en
  // cada lista
  list<int> L1,L2,*LL;
  list<int>::iterator p = L.begin();
  int n,n1=0,n2=0;
  while (p!=L.end()) {
    // para evitar duplicar codigo usamos un puntero
    // a listas `LL' que apunta a la lista elegida
    // al azar. 
    if (drand()>0.5) { LL = &L1; n1++; }
    else { LL = &L2; n2++; }
    LL->insert(LL->end(),*p);
    p = L.erase(p);
  }
  // Aca corta la recursividad. Si la lista es de
  // longitud 1 no hace falta aplicarle el `random_shuffle2'
  if (n1>1) random_shuffle2(L1);
  if (n2>1) random_shuffle2(L2);
  // Genera una lista random de dobles

  // RANDOM MERGE =============
  // Genera una lista de randoms de longitud n
  // igual a la de la lista. En aquellas posiciones
  // que tienen p<pc ponemos los elementos de la
  // lista L1 en los otros los de L2.
  // Primero hay que buscar `pc' tal que el numero
  // de elementos con p<pc sea exactamente `n1'

  // genera la lista de aleatorios
  n = n1+n2;
  list<double> P;
  for (int j=0; j<n; j++) P.insert(P.end(),drand());

  // Busca pc tal que haya n1 e
  // lems. con prob. <p y n2 con prob. > p
  // Esto se hace por busqueda binaria. (Esta parte
  // es O(n log n).
  double pa=0., pb=1.,pc;
  int n1c;
  list<double>::iterator r;
  while (true) {
    pc = (pa+pb)/2.;
    n1c = 0;
    r = P.begin();
    while (r!=P.end()) 
      if (*r++<pc) n1c++;
    if (n1c==n1) break;
    else if (n1c<n1) pa = pc;
    else pb = pc;
  }
  // Mirando la probabilidades en la lista
  // de aleatorios P si el elemento
  // es menor que pc lo insertamos en L1
  // y si no en L2. 
  r = P.begin();
  for (int j=0; j<n; j++) {
    LL = (*r++ < pc ? &L1 : &L2);
    list<int>::iterator q = LL->begin();
    assert(q!=LL->end());
    int w = *q;
    LL->erase(q);
    L.insert(L.end(),w);
  }
}

int main() {
  list<int> L;
  for (int j=0; j<=10; j++) {
    L.clear();
    randl(L,20,10.0);
    cout << "Antes de random_shuffle: ";
    printl(L);
    random_shuffle(L);
    cout << "Despues de random_shuffle: ";
    printl(L);
  }
}
